@@ -10,18 +10,20 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
+ * RCS: @(#) $Id: tclPort.h,v 1.7 1999/01/26 03:53:17 jingham Exp $
  */
 
 #ifndef _TCLPORT
 #define _TCLPORT
 
-#if defined(_Windows)
-#    include "tclWinPort.h"
-#elif defined(MAC_TCL)
-#    include "tclMacPort.h"
+#if defined(__WIN32__) || defined(_WIN32)
+#   include "../win/tclWinPort.h"
 #else
-#    include "tclUnixPort.h"
+#   if defined(MAC_TCL)
+#	include "tclMacPort.h"
+#    else
+#	include "../unix/tclUnixPort.h"
+#    endif
 #endif
 
 #endif /* _TCLPORT */
-
